# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring
# pylint: disable=line-too-long

from pathlib import Path

from fasthtml.fastapp import FastHTML, FastHTMLWithLiveReload, Request, StaticFiles, picolink
from fasthtml.common import Form, Label, Input, Br, Button, A, Html, serve, Body, Link, P, Body, Div, Span, Main, Title # pylint: disable=no-name-in-module
from starlette import status
from starlette.responses import Response, StreamingResponse
from starlette.exceptions import HTTPException
from pydantic import ValidationError

from .utils import load_database, save_database, generate_data
from .custom_types import TestDataBody, Config


class MaxSizePerRequestError(Exception):
    pass


class MinSizePerRequestError(Exception):
    pass


with open('./config.json') as file:  # parse_cli_arguments(sys.argv[1:])
    config = Config.model_validate_json(file.read())


app = FastHTMLWithLiveReload(hdrs=(
    (Link(rel='stylesheet', type='text/css', href='static/style.css'),)
))

app.mount(
    '/static',
    StaticFiles(directory=Path(__file__).parent.absolute() / 'static'),
    name='static'
)


@app.get('/')
async def homepage():
    return Main(
        P('Testdata Download'),
        P('Download a file of exact size. The file will only contain zeros.'),
        Div(
            Div(
                Form(
                    Div(
                        Label('API Key'), Br(),
                        Input(name='api_key'), Br()
                    ),
                    Div(
                        Label('Size'), Br(),
                        Input(name='size', placeholder='2MiB, 1GB, etc.'), Br()
                    ),
                    Button('Download'),
                    action='/zeros',  hx_boost="true"
                ),
                P(f'Max Download Size: {config.max_size / (1024 * 1024 * 1024):.02f} GiB'), Br(),  # nopep8
                cls='download'
            ),
            P('Contact: ', A('contact.testdata@krsnik.at', href='mailto:contact.testdata@krsnik.at'), id='contact')  # nopep8
        )
    )


@ app.get('/zeros')  # type: ignore
async def get_zeros(api_key: str, size: str, request: Request) -> StreamingResponse:
    # AJAX Requests cannot return a file, thus send a redirect.
    if 'HX-Request' in request.headers:
        return Response(
            content='<p>Hello</p>',
            headers={'HX-Redirect': f'/zeros/?api_key={api_key}&size={size}'})

    try:
        body = TestDataBody(api_key=api_key, size=size)  # type: ignore
    except ValidationError as err:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Invalid Format.'
        ) from err

    try:
        if body.api_key not in config.api_keys:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='Invalid API Key.'
            )

        if body.size < 0:
            raise MinSizePerRequestError

        if config.max_size < body.size:
            raise MaxSizePerRequestError

        db = load_database(config.database)
        if config.max_data <= db['data-used'] + body.size:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail='Service not available.'
            )

        db['data-used'] += body.size
        save_database(config.database, db)

        return StreamingResponse(
            status_code=status.HTTP_200_OK,
            content=generate_data(body.size, config.buffer_size),
            media_type='application/octet-stream',
            headers={
                'Content-Length': str(body.size),
            }
        )

    except MinSizePerRequestError as err:
        raise HTTPException(
            status_code=status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
            detail='Size has to be not-negative.'
        ) from err
    except MaxSizePerRequestError as err:
        raise HTTPException(
            status_code=status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
            detail=f'Exceeded max size per request of {config.max_size} Bytes.'
        ) from err
