import json
import asyncio
from typing import AsyncGenerator


def convert_to_bytes(size: int | str) -> int:
    if isinstance(size, int):
        return size

    try:
        return int(size)
    except ValueError as err:
        units = {
            'TB': 1000 ** 4, 'TiB': 1024 ** 4,
            'GB': 1000 ** 3, 'GiB': 1024 ** 3,
            'MB': 1000 ** 2, 'MiB': 1024 ** 2,
            'KB': 1000, 'KiB': 1024,
            'B': 1
        }

        for unit, value in units.items():
            if size.endswith(unit):
                return int(float(size.removesuffix(unit)) * value)

        raise ValueError(
            'Invalid format. Expected integer or float ending with a data unit (B, KB, MiB,...).'
        ) from err


async def generate_data(size: int, buffer_size: int = 4 * 1024) -> AsyncGenerator[bytes, None]:
    size_left = size

    # https://github.com/tiangolo/fastapi/issues/5183
    # https://github.com/encode/starlette/discussions/1776#discussioncomment-3207518

    try:
        while size_left > buffer_size:
            size_left -= buffer_size
            yield b'\0' * buffer_size
            await asyncio.sleep(0)

        yield b'\0' * size_left
        await asyncio.sleep(0)
    except asyncio.CancelledError as err:
        raise GeneratorExit from err


def load_database(path: str) -> dict:
    with open(path, 'r', encoding='utf-8') as file:
        return json.load(file)


def save_database(path: str, database: dict) -> None:
    with open(path, 'w', encoding='utf-8') as file:
        json.dump(database, file, indent=2)
